;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2024 Rutger (regtur) van Beusekom <rutger@dezyne.org>
;;; Copyright © 2019, 2021, 2022, 2023 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
;;; Copyright © 2019 Rob Wieringa <rma.wieringa@gmail.com>
;;; Copyright © 2025 Paul Hoogendijk <paul@dezyne.org>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
;;;
;;; Commentary:
;;;

(define-module (dzn ast recursive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)

  #:use-module (ice-9 match)

  #:use-module (dzn ast ast)
  #:use-module (dzn ast normalize)
  #:use-module (dzn ast)
  #:use-module (dzn misc)
  #:export (recursive:annotate))

(define-method (set-context-called (o <behavior>))
  (define (call* o)
    (tree:collect o
                  (is? <call>)
                  #:stop? (disjoin (is? <behavior>)
                                   (is? <functions>)
                                   (is? <statement>)
                                   (is? <arguments>)
                                   (is? <expression>))))
  (define (function* c*)
    (delete-duplicates (filter (is? <function>) (map .function c*)) eq?))
  (define (called-function* f)
    (function* (call* (.statement f))))
  (define (star f*)
    (let loop ((f* f*))
      (let* ((step-f* (append f* (append-map called-function* f*)))
             (step-f* (delete-duplicates step-f* eq?)))
        (if (eq? (length step-f*) (length f*)) f*
            (loop step-f*)))))

  (let* ((calls (call* o))
         (declarative-called (star (function* (filter ast:declarative? calls))))
         (imperative-called (star (function* (filter ast:imperative? calls)))))
    (define (mark-context-called f)
      (let* ((declarative-called? (and (memq f declarative-called) #t))
             (imperative-called? (and (memq f imperative-called) #t))
             (context (cond ((and declarative-called? imperative-called?) 'both)
                            (declarative-called? 'declarative)
                            (imperative-called? 'imperative)
                            (else #f))))
        (clone f #:context-called context)))
    (let* ((functions (.functions o))
           (function-list (.elements functions))
           (function-list (map mark-context-called function-list))
           (functions (clone functions #:elements function-list)))
      (clone o #:functions functions))))

(define (mark-recursive f)
  (if (not (ast:recursive? f)) f
      (clone f #:recursive? #t)))

(define-method (set-recursive (o <behavior>))
  (let* ((functions (.functions o))
         (function-list (.elements functions))
         (function-list (map mark-recursive function-list))
         (functions (clone functions #:elements function-list)))
    (clone o #:functions functions)))

(define (recursive:annotate o)
  (match o
    (($ <behavior>)
     ((compose set-context-called set-recursive) o))
    ((? (%normalize:short-circuit?))
     o)
    ((or ($ <interface>) ($ <component>))
     (clone o #:behavior (and=> (.behavior o)
                                (compose set-context-called set-recursive))))
    (($ <function>)
     (mark-recursive o))
    ((? (is? <ast>))
     (tree:shallow-map recursive:annotate o))
    (_
     o)))
